/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

 .bk.bm-tool-loading {
    overflow: hidden;
  }
  .bk.bm-tool-loading:before {
    position: absolute;
    height: 100%;
    width: 100%;
    content: '';
    z-index: 1000;
    background-color: rgb(255, 255, 255, 0.75);
    border-color: lightgray;
    background-repeat: no-repeat;
    background-position: center;
    background-size: auto 50%;
    border-width: 1px;
    cursor: wait;
  }
  .bk.bm-tool-loading.arcs:hover:before {
    content: 'Please select a Query from the Select menu above.';
    font: x-large Arial, sans-serif;
    color: black;
    cursor: progress;
    display: flex;
    justify-content: center;
    align-items: center;
  }
